import concurrent.futures
from threading import BoundedSemaphore

from apipulse_python.sdk_logger import logger

from .http.http_connection import ApiPulseHttpConnection


class ApiPulseDataSyncService:
    def __init__(self, apipulse_http_connection):
        self.__apipulse_http_connection: ApiPulseHttpConnection = apipulse_http_connection
        self.__drop_count = 0
        self.__executor_service = None
        self.__semaphore = BoundedSemaphore(5)

    def init(self):
        try:
            self.__executor_service = concurrent.futures.ThreadPoolExecutor(
                max_workers=2, thread_name_prefix="sl-sync-"
            )
        except Exception as e:
            logger.error("Error in ApipulseDataSyncService.init", exc_info=e)
            return False
        return True

    def sync_data(self, api_sample):
        try:
            if api_sample is None:
                return

            def send_samples():
                samples = [api_sample]
                result = self.__apipulse_http_connection.send_samples(samples)
                if not result:
                    self.__drop_count += 1
                    logger.info(f"Dropped {self.__drop_count} samples")
                self.__semaphore.release()

            if self.__semaphore.acquire(blocking=False):
                self.__executor_service.submit(send_samples)
            else:
                self.__drop_count += 1
                logger.info(f"Dropped {self.__drop_count} samples")
        except Exception as e:
            logger.error("Error in ApipulseDataSyncService.sync_data", exc_info=e)



class HttpResponse:
    def __init__(
        self,
        headers,
        status_code,
        body_raw=None,
        body=None,
    ):
        self.headers = headers
        self.status_code = status_code
        self.body_raw = body_raw
        self.body = body



from abc import ABC, abstractmethod

from .config_update_listener import ConfigUpdateListener


class ConfigManager(ABC):
    @abstractmethod
    def init(self) -> bool:
        pass

    @abstractmethod
    def subscribe_to_updates(self, config_update_listener: ConfigUpdateListener) -> bool:
        pass

    @abstractmethod
    def shutdown(self) -> bool:
        pass



from typing import Dict, List


def get_masked_headers(headers: Dict[str, str], mask_headers: List[str]):
    if not headers or not mask_headers or len(mask_headers) == 0:
        return headers

    header_names = headers.keys()

    for header_name in header_names:
        if header_name.casefold() in (name.casefold() for name in mask_headers):
            headers[header_name] = "🔒MASKED🔒"
    return headers



import logging
import sys
from typing import List, Optional, Union

from apipulse_python.core.model.sdk_options import SdkOptions

from .auto_configuration import ApiPulseAutoConfiguration

options: Optional[SdkOptions] = None
initialized = False
auto_configuration: Optional[ApiPulseAutoConfiguration] = None



def init(
    url: str = "",
    application_name: str = "",
    environment: str = "",
    auth_key: str = None,
    logging_enabled: bool = False,
    log_level: Union[str, int] = logging.FATAL,
    mask_headers: Optional[List[str]] = None,
    capture: str = None,
    # extra 
    partner_id: str = None,
    team_name: str = None,
     
):
    """
    Initialize Apipulse SDK. It automatically samples traffic and syncs with Apipulse servers.


    :param str url: The Apipulse url for your org, provided by Apipulse team.
        Ex: https://example.apipulse.dev

    :param str application_name: Project or Application name. Ex: Demo-Service

    :param str auth_key: Apipulse auth key, provided by Apipulse team.

    :param str environment: Your application environment. Ex: stage, prod, alpha, etc.

    :param bool logging_enabled: Enable Apipulse SDK logging

    :param str log_level: Apipulse SDK logging level. Logging depends on `logging_enabled ` argument.

    :param list(int) mask_headers: Headers names (case in-sensitive) that will be masked at the SDK level itself.
        Masking will happen locally before data is sent to Apipulse servers. Ex: ["cookie", "x-auth"]

    :param str capture: TEST MODE
    """
    global options, initialized, auto_configuration

    # sdk fails in django-admin commands. we only allow sdk to run if runserver django admin command was used
    if len(sys.argv) >= 2 and sys.argv[0].endswith("manage.py") and sys.argv[1] != "runserver":
        return

    if not initialized:
        initialized = True

        opts: SdkOptions = SdkOptions(
            url, application_name, auth_key, environment, logging_enabled, log_level, mask_headers, capture, partner_id, team_name
        )
        opts.sanitize()
        is_options_valid = opts.validate()

        if not is_options_valid:
            return

        options = opts

        auto_configuration = ApiPulseAutoConfiguration(opts=opts)
        auto_configuration.init()

        print("Apipulse Initialized !")



from typing import Optional

import apipulse_python.core.model as model

from ..constant import HttpMethod


class ObservedApi:
    def __init__(self, uri: str, method: HttpMethod):
        self.uri: model.URI = model.URI.get_non_templated_uri(uri)
        self.method: HttpMethod = method

    def matches(self, api_config: Optional[model.ApiConfig]) -> bool:
        if api_config is None:
            return False
        if self.method.value != api_config.method:
            return False
        return self.uri == api_config.uri
    




import logging
import sys
from typing import Union

logger = logging.getLogger("apipulse_python")
logger.propagate = False

stream_handler = logging.StreamHandler(stream=sys.stdout)
stream_handler.setFormatter(logging.Formatter("%(asctime)s %(levelname)s %(name)s %(threadName)s %(message)s"))
logger.addHandler(stream_handler)


def set_logging_config(enabled: bool, level: Union[int, str]):
    global logger
    if not enabled:
        # effectively disables logging
        logger.setLevel(logging.CRITICAL + 1)
    else:
        logger.setLevel(level)



import requests
from django.urls import get_resolver
from django.urls import URLResolver, URLPattern
from . import constants
from .contract import ServiceApiDetails, EndpointInfo
from .utils import simplify_regex_pattern
import json
# from apipulse_python.core.model.sdk_options import SdkOptions
# from apipulse_python.core.shortlook_sdk import options as OPTIONS



def collect_urls(urllist, prefix=''):
    url_info = []
    for entry in urllist:
        if isinstance(entry, URLPattern):
            pattern = prefix + entry.pattern.regex.pattern
            view = entry.callback
            methods = getattr(view, 'methods', ['GET'])  # Default to ['GET'] if not specified
            url_info.append({'url': pattern, 'methods': methods})
        elif isinstance(entry, URLResolver):
            url_info.extend(collect_urls(entry.url_patterns, prefix + entry.pattern.regex.pattern))
    return url_info

    
def collect_and_send_urls():
    url_info = collect_urls(get_resolver().url_patterns)
    endpoint_url = constants.DATA_INGESTION_URL
    endpoint_list = list()
    for info_dict in url_info:
        url = info_dict["url"]
        methods = info_dict["methods"]
        # create endpoint
        for method in methods:
            url = simplify_regex_pattern(url)
            endpoint_list.append(EndpointInfo(pattern=url,method=method))
    # done endpoint list      
    # print("endpointlist ", endpoint_list) 
    serviceApiDetails = ServiceApiDetails(controllerVsApis={"root": endpoint_list})  
    # print("serviceApiDetails", serviceApiDetails) 
    data = serviceApiDetails.model_dump()
    data["serviceName"] = constants.HEADERS["X-SERVICE-NAME"]
    data["team"] = constants.HEADERS["X-TEAM-NAME"]
    data["env"] = constants.HEADERS["X-ENV-NAME"]
    print("data collected for API : ",data) 
    try:
        response = requests.post(endpoint_url, headers=constants.HEADERS, json=data)
        print(f"URLs sent. Response: {response.status_code} {response.text} {response.json()}")
    except Exception as e:
        print(f"Error sending URLs: {e} ")
        






import re

from apipulse_python.core.http import HttpRequest, HttpResponse, RequestResponseContext
from apipulse_python.core.apipulse_sdk import auto_configuration
from apipulse_python.api_pulse_middleware.collect_urls import collect_and_send_urls


class ApiPulseDjangoMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response
        collect_and_send_urls()
        
    def __call__(self, request):
        # TODO handle async

        # sometimes host contains port too
        host, port = self.get_host_port(request.get_host(), request.get_port())

        http_request = HttpRequest(
            raw_uri=request.path,
            hostname=host,
            scheme=request.scheme,
            port=port,
            method=request.method,
            headers=dict(request.headers.items()),
            params=dict(request.GET.lists()),
            body_raw=request.body,
        )

        ctx = RequestResponseContext(request=http_request)
        if auto_configuration and auto_configuration.filter:
            return auto_configuration.filter.process(ctx, self.wrap_get_response, request)
        else:
            return self.get_response(request)


    def wrap_get_response(self, *args, **kwargs):
        # response is the raw response of the next function. this is framework dependent.
        # return type can be None because maybe the next function doesn't return anything
        # regardless of the return type, we return this from the middleware so that it can be used in the chain
        #
        # http_response is parsed object of type apipulse_python.core.http.HttpResponse created from framework_response
        # this will never be None but it's properties can be.
        args = args if args is not None else []
        kwargs = kwargs if kwargs is not None else {}
        response = self.get_response(*args, **kwargs)
        if hasattr(response, 'content'):
            return response, HttpResponse(
                headers=self.resolve_response_header(response),
                status_code=response.status_code,
                body_raw=response.content,
            )
        else:
            return response, HttpResponse(
                headers=self.resolve_response_header(response),
                status_code=response.status_code,
                body_raw=bytearray(),
            )
    @staticmethod
    def resolve_response_header(response):
        if response is None:
            return dict()
        if hasattr(response, 'headers'):
            return dict(response.headers.items())
        if hasattr(response, '_headers'):
            header_dict = {}
            headers = response._headers
            for header_name in headers.keys():
                if type(headers[header_name]) == str:
                    header_dict[header_name] = headers[header_name]
                elif type(headers[header_name]) == tuple and len(headers[header_name]) >=2:
                    header_dict[header_name] = headers[header_name][1]

            return header_dict

    @staticmethod
    def get_host_port(host: str, port: int):
        pattern = r"(.+?):(\d+)"
        match = re.match(pattern, host)
        if match:
            p_host = match.group(1)
            p_port = match.group(2)
            return p_host, p_port
        else:
            return host, port



class ApiSample:
    def __init__(
        self,
        raw_uri,
        application_name,
        host_name,
        port,
        scheme,
        method,
        parameters,
        request_headers,
        response_headers,
        status_code,
        request_payload,
        response_payload,
        uncaught_exception_message,
        payload_capture_attempted,
        request_payload_capture_attempted,
        response_payload_capture_attempted,
        latency,
        uri
    ):
        self.raw_uri = raw_uri
        self.application_name = application_name
        self.host_name = host_name
        self.port = port
        self.scheme = scheme
        self.method = method
        self.parameters = parameters
        self.request_headers = request_headers
        self.response_headers = response_headers
        self.status_code = status_code
        self.request_payload = request_payload
        self.response_payload = response_payload
        self.uncaught_exception_message = uncaught_exception_message
        self.payload_capture_attempted = payload_capture_attempted
        self.request_payload_capture_attempted = request_payload_capture_attempted
        self.response_payload_capture_attempted = response_payload_capture_attempted
        self.latency = latency
        self.uri = uri

    def __str__(self):
        return (
            f"ApiSample(raw_uri={self.raw_uri}, application_name={self.application_name}, "
            f"host_name={self.host_name}, port={self.port}, scheme={self.scheme}, method={self.method}, "
            f"parameters={self.parameters}, request_headers={self.request_headers}, "
            f"response_headers={self.response_headers}, status_code={self.status_code}, "
            f"request_payload={self.request_payload}, response_payload={self.response_payload}, "
            f"uncaught_exception_message={self.uncaught_exception_message}, "
            f"payload_capture_attempted={self.payload_capture_attempted}, "
            f"request_payload_capture_attempted={self.request_payload_capture_attempted}, "
            f"response_payload_capture_attempted={self.response_payload_capture_attempted}, "
            f"latency={self.latency}, "
            f"uri={str(self.uri)})"
        )


def __eq__(self, other):
    if not isinstance(other, ApiSample):
        return False
    return (
        self.raw_uri == other.raw_uri
        and self.application_name == other.application_name
        and self.host_name == other.host_name
        and self.port == other.port
        and self.scheme == other.scheme
        and self.method == other.method
        and self.parameters == other.parameters
        and self.request_headers == other.request_headers
        and self.response_headers == other.response_headers
        and self.status_code == other.status_code
        and self.request_payload == other.request_payload
        and self.response_payload == other.response_payload
        and self.uncaught_exception_message == other.uncaught_exception_message
        and self.payload_capture_attempted == other.payload_capture_attempted
        and self.request_payload_capture_attempted == other.request_payload_capture_attempted
        and self.response_payload_capture_attempted == other.response_payload_capture_attempted
        and self.latency == other.latency
    )



from abc import ABC, abstractmethod

from apipulse_python.core.model.agent_config import AgentConfig


class ConfigUpdateListener(ABC):
    @abstractmethod
    def on_successful_config_update(self, agent_config: AgentConfig):
        pass

    @abstractmethod
    def on_erroneous_config_update(self):
        pass



import requests
from django.urls import get_resolver
from django.urls import URLResolver, URLPattern
from . import constants
from .contract import ServiceApiDetails, EndpointInfo
from .utils import simplify_regex_pattern
import json

class UrlCollectionMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response
        self.collect_and_send_urls()

    def collect_urls(self,urllist, prefix=''):
        url_info = []
        for entry in urllist:
            if isinstance(entry, URLPattern):
                pattern = prefix + entry.pattern.regex.pattern
                view = entry.callback
                methods = getattr(view, 'methods', ['GET'])  # Default to ['GET'] if not specified
                url_info.append({'url': pattern, 'methods': methods})
            elif isinstance(entry, URLResolver):
                url_info.extend(self.collect_urls(entry.url_patterns, prefix + entry.pattern.regex.pattern))
        return url_info

    
    def collect_and_send_urls(self):
        url_info = self.collect_urls(get_resolver().url_patterns)
        endpoint_url = constants.API_PULSE_URL  # Replace with your actual endpoint
        endpoint_list = list()
        for info_dict in url_info:
            url = info_dict["url"]
            methods = info_dict["methods"]
            # create endpoint
            for method in methods:
                url = simplify_regex_pattern(url)
                endpoint_list.append(EndpointInfo(pattern=url,method=method))
        # done endpoint list      
        # print("endpointlist ", endpoint_list) 
        serviceApiDetails = ServiceApiDetails(controllerVsApis={"root": endpoint_list})  
        # print("serviceApiDetails", serviceApiDetails) 
        data = serviceApiDetails.model_dump()
        data["serviceName"] = constants.HEADERS["X-SERVICE-NAME"]
        data["team"] = constants.HEADERS["X-TEAM-NAME"]
        data["env"] = constants.HEADERS["X-ENV-NAME"]
        print("data collected for API : ",data) 
        try:
            response = requests.post(endpoint_url, headers=constants.HEADERS, json=data)
            print(f"URLs sent. Response: {response.status_code} {response.text} {response.json()}")
        except requests.RequestException as e:
            print(f"Error sending URLs: {e} ")

    def __call__(self, request):
        response = self.get_response(request)
        return response




from apipulse_python.core.model.agent_config import AgentConfig
from apipulse_python.core.model.api_config import ApiConfig
from apipulse_python.core.model.uri import URI


def is_agent_config_valid(agent_config: AgentConfig) -> bool:
    if agent_config is None:
        return False

    if agent_config.buffer_sync_freq_in_sec is None:
        return False

    if agent_config.config_fetch_freq_in_sec is None:
        return False

    if agent_config.capture_api_sample is None:
        return False

    if agent_config.discovery_buffer_size is None:
        return False

    if agent_config.discovery_buffer_size_per_api is None:
        return False

    if agent_config.registered_api_configs is not None:
        for api_config in agent_config.registered_api_configs:
            if not is_api_config_valid(api_config):
                return False

    if agent_config.black_list_rules is not None:
        for black_list_rule in agent_config.black_list_rules:
            if not is_black_list_rule_valid(black_list_rule):
                return False

    return True


def is_api_config_valid(api_config: ApiConfig) -> bool:
    if api_config is None:
        return False

    if not is_uri_valid(api_config.uri):
        return False

    if api_config.method is None:
        return False

    if api_config.buffer_size is None:
        return False

    if api_config.capture_sample_request is None:
        return False

    if api_config.capture_sample_response is None:
        return False

    return True


def is_uri_valid(uri: URI) -> bool:
    if uri is None:
        return False

    if uri.uri_path is None:
        return False

    if len(uri.uri_path) == 0:
        return False

    return True


def is_black_list_rule_valid(black_list_rule) -> bool:
    if black_list_rule is None:
        return False

    if not black_list_rule.is_valid():
        return False

    return True



from enum import Enum


class HttpMethod(str, Enum):
    GET = "GET"
    HEAD = "HEAD"
    POST = "POST"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"
    OPTIONS = "OPTIONS"
    TRACE = "TRACE"



import queue
from typing import Optional

from ..model import ApiSample
from .buffer import Buffer


class SimpleBuffer(Buffer):
    def __init__(self, max_size):
        self.__sample_queue = queue.Queue(max_size)

    def offer(self, sample: ApiSample) -> bool:
        try:
            self.__sample_queue.put(sample, block=False)
        except queue.Full:
            pass
        return True

    def can_offer(self):
        return not self.__sample_queue.full()

    def poll(self) -> Optional[ApiSample]:
        try:
            return self.__sample_queue.get(block=False)
        except queue.Empty:
            return None

    def clear(self):
        self.__sample_queue.queue.clear()

    def size(self):
        return self.__sample_queue.qsize()

    def __str__(self):
        return f"{str(self.__sample_queue)}"



from .agent_config import AgentConfig
from .api_config import ApiConfig
from .api_sample import ApiSample
from .black_list_rule import BlackListRule
from .observed_api import ObservedApi
from .sdk_options import SdkOptions
from .uri import URI

__all__ = ["AgentConfig", "ApiConfig", "SdkOptions", "BlackListRule", "URI", "ObservedApi", "ApiSample"]



from apipulse_python.core.api_processor import ApiProcessor
from apipulse_python.core.buffer.api_buffer_key import ApiBufferKey
from apipulse_python.core.http.context import RequestResponseContext
from apipulse_python.core.util.filter_utils import get_observed_api_from_request
from apipulse_python.sdk_logger import logger
import time

class AlwaysCaptureApipulseFilter:
    def __init__(self, api_processor, application_name):
        self.__api_processor: ApiProcessor = api_processor
        self.__application_name: str = application_name

    def process(self, ctx: RequestResponseContext, next_fn, *args, **kwargs):
        start_time = time.time()
        framework_response, http_response = next_fn(*args, **kwargs)
        end_time = time.time()
        try:
            observed_api = get_observed_api_from_request(ctx.request)

            ctx.application_name = self.__application_name
            ctx.observed_api = observed_api
            ctx.api_buffer_key = ApiBufferKey.get_api_buffer_key_from_observed_api(observed_api)
            ctx.latency = int((end_time - start_time) * 1000)
            self.__api_processor.process_always_capture(ctx, http_response)
        except Exception as e:
            logger.error("Error in Apipulse.process", exc_info=e)
        return framework_response



import pprint
import re

def simplify_regex_pattern(pattern: str) -> str:
    # Remove start (^) and end ($ or \\Z) of string regex symbols
    pattern = pattern.replace('^', '').replace('$', '').replace('\\Z', '')

    # Replace named regex groups with a more readable format using {}
    pattern = re.sub(r'\(\?P<(\w+)>[^)]+\)', r'{\1}', pattern)

    # Replace unnamed URL parameters (e.g., (\d+)) with a generic placeholder
    pattern = re.sub(r'\([^\)]*\)', '{param}', pattern)

    # Simplify common regex patterns
    pattern = pattern.replace('\\.', '.').replace('\\-', '-').replace('\\/', '/')

    # Handle optional parts in the URL
    pattern = pattern.replace('/?', '/')

    # Additional simplifications can be added here

    return pattern



def test():
    pattern1 = '^admin/^auth/user/^(?P<object_id>.+)/change/\\Z'
    print(simplify_regex_pattern(pattern1))
    
    pattern2 = '^browser/^number/(?P<number>[0-9]+)/\\Z'
    print(simplify_regex_pattern(pattern2))

if __name__=='__main__':
    # Example usage
    test()



import json
from typing import List, Optional
from urllib.parse import urlencode
from urllib.request import Request, urlopen

from apipulse_python.sdk_logger import logger
from apipulse_python.sdk_version import APIPULSE_SDK_VERSION
from apipulse_python.core.util.common_utils import get_headers

from ..model import AgentConfig, ApiSample


class ApiPulseHttpConnection:
    def __init__(self, ct_url, auth_key, environment, capture):
        self.__base_url = ct_url
        self.__auth_key = auth_key
        self.__environment = environment
        self.__always_capture = capture == "always"
        self.__headers = {"Accept": "application/json", "Connection": "close", **APIPULSE_SDK_VERSION}
        self.__headers.update(get_headers())

        if auth_key:
            self.__headers["authKey"] = self.__auth_key

        if environment and len(environment) > 0:
            self.__headers["environment"] = self.__environment

        if self.__always_capture:
            self.__headers["apipulse-capture"] = "always"

    def agent_config(self, agent_id, app_name) -> Optional[AgentConfig]:
        try:
            if len(self.__base_url) == 0:
                logger.error("Apipulse base url is empty")
                return None

            query_params = {
                "agentId": agent_id,
                "appName": app_name,
            }

            url = self.__base_url + "/api/v1/mirror/agent-config"

            if query_params:
                url = url + "?" + urlencode(query_params)

            self.__headers.update(get_headers())
            http_request = Request(url, method="GET", headers=self.__headers)

            with urlopen(http_request) as response:
                response_obj = json.load(response)
                logger.info(">>> response_obj ",response_obj)
                return AgentConfig(
                    buffer_sync_freq_in_sec=response_obj["bufferSyncFreqInSec"],
                    capture_api_sample=response_obj["captureApiSample"],
                    config_fetch_freq_in_sec=response_obj["configFetchFreqInSec"],
                    registered_api_configs=response_obj["registeredApiConfigs"],
                    timestamp=response_obj["timestamp"],
                    discovery_buffer_size=response_obj["discoveryBufferSize"],
                    discovery_buffer_size_per_api=response_obj["discoveryBufferSizePerApi"],
                    black_list_rules=response_obj["blackListRules"],
                )
        except Exception as e:
            logger.error("Error while fetching agent config", exc_info=e)
            return None

    def send_data_to_api_pulse(self,url,headers, data):
        ...
        # TODO AMIT changes
        import requests,json
        try:
            # modify the data 
            from apipulse_python.api_pulse_middleware.contract import ApiOwner, ApiSampleWrapper
            from apipulse_python.api_pulse_middleware.constants import HEADERS
            apiOwner = ApiOwner(env=HEADERS["X-ENV-NAME"], team = HEADERS["X-TEAM-NAME"], serviceName= HEADERS["X-SERVICE-NAME"] )
            data = json.loads(data)
            apiSampleWrapper = ApiSampleWrapper(apiOwner=apiOwner, apiSamples=data)
            print(f"Apisamplewrapper = {apiSampleWrapper}")
            json_data = apiSampleWrapper.model_dump()
            # json_data = json.loads(json_data)
            custom_header = dict()
            custom_header.update(HEADERS)
            response = requests.post(url=url,headers=custom_header, json=json_data)
            print(f"ApiPulse Ingestion apiSampleWrapper: {response.status_code} {response.text} {response.json()}")
        except Exception as e:
            print(f"Error sending URLs: {e} ")
    
    def send_samples(self, contents: List[ApiSample]):
        logger.debug(f"sending samples: [{','.join(map(lambda s: str(s), contents))}]")
        try:
            if len(self.__base_url) == 0:
                logger.error("Apipulse base url is empty")
                return None

            json_samples = []
            for content in contents:
                json_samples.append(to_json(content))
            data = json.dumps(json_samples, ensure_ascii=False)

            url = self.__base_url + "/api/v1/mirror/data-ingestion/api-sample"

            if self.__always_capture:
                url = self.__base_url + "/api/v1/mirror/data-ingestion/api-sample"
                
            self.__headers.update(get_headers())
            http_request = Request(url, method="POST", headers=self.__headers, data=data.encode("utf-8"))
            # http_request.add_header("Content-Type", "application/json")
            self.send_data_to_api_pulse(url,self.__headers, data)
            # with urlopen(http_request) as response:
            #     print(f"Response = {response}")
            #     if not response.status == 200:
            #         logger.error(f"Send Sample Request failed, API returned {response.status}")
            #         return False

            return True
        except Exception as e:
            logger.error("Error while sending samples", exc_info=e)
            return False


def to_json(sample: ApiSample):
    data = {
        "rawUri": sample.raw_uri,
        "applicationName": sample.application_name,
        "hostName": sample.host_name,
        "port": int(sample.port) if sample.port and isinstance(sample.port, str) else sample.port,
        "scheme": sample.scheme,
        "method": sample.method,
        "statusCode": int(sample.status_code)
        if sample.status_code and isinstance(sample.status_code, str)
        else sample.status_code,
        "requestPayload": sample.request_payload,
        "responsePayload": sample.response_payload,
        "uncaughtExceptionMessage": sample.uncaught_exception_message,
        "payloadCaptureAttempted": sample.payload_capture_attempted,
        "requestPayloadCaptureAttempted": sample.request_payload_capture_attempted,
        "responsePayloadCaptureAttempted": sample.response_payload_capture_attempted,
        "latency": sample.latency,
        "uri":{
        "uriPath": sample.uri.uri_path,
        "hasPathVariable": sample.uri.has_path_variable
        }
    }

    params = {}
    req_headers = {}
    res_headers = {}

    if sample.parameters:
        keys = list(sample.parameters.keys())
        for key in keys:
            params[key] = sample.parameters.get(key)
    if sample.request_headers:
        keys = list(sample.request_headers.keys())
        for key in keys:
            req_headers[key] = sample.request_headers.get(key)
    if sample.response_headers:
        keys = list(sample.response_headers.keys())
        for key in keys:
            res_headers[key] = sample.response_headers.get(key)

    data["parameters"] = params
    data["requestHeaders"] = req_headers
    data["responseHeaders"] = res_headers

    return data



from typing import List, Optional

from ..constant import HttpMethod
from ..http.http_request import HttpRequest
from ..model import AgentConfig, ApiConfig, BlackListRule, ObservedApi


def get_api_config(observed_api: ObservedApi, agent_config: AgentConfig) -> Optional[ApiConfig]:
    if not len(agent_config.registered_api_configs) > 0:
        return None
    registered_api_configs = agent_config.registered_api_configs
    for api_config in registered_api_configs:
        if observed_api.matches(api_config):
            return api_config

    return None


def get_observed_api_from_request(request: HttpRequest):
    method = HttpMethod(request.method)
    return ObservedApi(request.raw_uri, method)


def is_blacklisted_api(observed_api: ObservedApi, agent_config: AgentConfig) -> bool:
    try:
        if not agent_config or not agent_config.black_list_rules:
            return False
        blacklist_rules: List[BlackListRule] = agent_config.black_list_rules
        for blacklist_rule in blacklist_rules:
            if blacklist_rule.matches_uri(observed_api.uri, observed_api.method):
                return True
    except Exception:
        pass
    return False



import time

from apipulse_python.sdk_logger import logger

from .always_capture_sync import ApiPulseDataSyncService
from .buffer.buffer_manager_worker import BufferManagerWorker
from .buffer.discovered_buffer_manager import DiscoveredApiBufferManager
from .buffer.registered_buffer_manager import RegisteredApiBufferManager
from .http import RequestResponseContext
from .model.api_sample import ApiSample
from .util.masking_utils import get_masked_headers


class ApiProcessor:
    def __init__(
        self,
        mask_headers,
        registered_api_buffer_manager=None,
        discovered_api_buffer_manager=None,
        data_sync_service=None,
    ):
        self.__mask_headers = mask_headers
        self.__data_sync_service: ApiPulseDataSyncService = data_sync_service
        self.__registered_api_buffer_manager: RegisteredApiBufferManager = registered_api_buffer_manager
        self.__discovered_api_buffer_manager: DiscoveredApiBufferManager = discovered_api_buffer_manager

    def process_always_capture(self, ctx: RequestResponseContext, http_response):
        ctx.response = http_response
        ctx.request.body = str(ctx.request.body_raw, "utf-8")
        ctx.response.body = str(ctx.response.body_raw, "utf-8")
        ctx.payload_capture_attempted = True
        ctx.request_payload_capture_attempted = True
        ctx.response_payload_capture_attempted = True

        self.__try_offering_always_capture(ctx)

    def __try_offering_always_capture(self, ctx: RequestResponseContext):
        try:
            logger.debug("ApiProcessor.__try_offering_always_capture")
            api_sample = self.__get_sample_from_context(ctx)
            self.__data_sync_service.sync_data(api_sample)
        except Exception as e:
            logger.error("Error in ApiProcessor.__try_offering_always_capture", exc_info=e)

    def process_discovered_api(self, ctx: RequestResponseContext, http_response):
        worker = self.__discovered_api_buffer_manager.worker
        if not worker:
            logger.error("__discovered_api_buffer_manager.worker is None inside process_discovered_api")
            return

        can_offer = worker.can_offer(ctx.api_buffer_key)
        ctx.payload_capture_attempted = False
        ctx.response = http_response
        ctx.request.body = None
        ctx.response.body = None

        if can_offer and worker:
            self.__try_offering(ctx, worker)

    def process_registered_api(self, ctx: RequestResponseContext, http_response):
        worker = self.__registered_api_buffer_manager.worker
        if not worker:
            logger.error("__registered_api_buffer_manager.worker is None inside process_registered_api")
            return

        can_offer = True
        ctx.response = http_response

        request_payload_capture_attempted = self.__should_capture_request(ctx)
        if not request_payload_capture_attempted:
            ctx.request.body = None
        elif ctx.request.body_raw:
            ctx.request.body = str(ctx.request.body_raw, "utf-8")

        response_payload_capture_attempted = self.__should_capture_response(ctx)
        if not response_payload_capture_attempted:
            ctx.response.body = None
        elif ctx.response.body_raw:
            ctx.response.body = str(ctx.response.body_raw, "utf-8")

        ctx.payload_capture_attempted = True
        ctx.request_payload_capture_attempted = request_payload_capture_attempted
        ctx.response_payload_capture_attempted = response_payload_capture_attempted

        if can_offer and worker:
            self.__try_offering(ctx, worker)

    def __try_offering(self, ctx: RequestResponseContext, worker: BufferManagerWorker):
        try:
            logger.debug("ApiProcessor.__try_offering")
            api_sample = self.__get_sample_from_context(ctx)
            worker.offer(ctx.api_buffer_key, api_sample)
        except Exception as e:
            logger.error("Error in ApiProcessor.__try_offering", exc_info=e)

    @staticmethod
    def __should_capture_request(ctx: RequestResponseContext) -> bool:
        return ctx and ctx.api_config and ctx.api_config.capture_sample_request

    @staticmethod
    def __should_capture_response(ctx: RequestResponseContext) -> bool:
        return ctx and ctx.api_config and ctx.api_config.capture_sample_response

    def __get_sample_from_context(self, ctx: RequestResponseContext):
        api_sample = ApiSample(
            raw_uri=ctx.request and ctx.request.raw_uri,
            method=ctx.request and ctx.request.method,
            application_name=ctx.application_name,
            host_name=ctx.request.hostname,
            port=ctx.request.port,
            scheme=ctx.request and ctx.request.scheme,
            parameters=ctx.request and ctx.request.params,
            request_headers=ctx.request and get_masked_headers(ctx.request.headers, self.__mask_headers),
            response_headers=ctx.response and get_masked_headers(ctx.response.headers, self.__mask_headers),
            status_code=ctx.response and ctx.response.status_code,
            request_payload=ctx.request and ctx.request.body,
            response_payload=ctx.response and ctx.response.body,
            uncaught_exception_message=None,
            payload_capture_attempted=ctx.payload_capture_attempted,
            request_payload_capture_attempted=ctx.request_payload_capture_attempted,
            response_payload_capture_attempted=ctx.response_payload_capture_attempted,
            latency=ctx.latency,
            uri = ctx.observed_api.uri
        )

        if ctx.api_config:
            api_sample.method = ctx.api_config.method
        else:
            api_sample.method = ctx.observed_api.method

        return api_sample



from abc import ABC, abstractmethod
from typing import Optional

from ..config.config_manager import ConfigManager
from ..config.config_update_listener import ConfigUpdateListener
from ..model import AgentConfig
from .buffer_manager_worker import BufferManagerWorker
from .no_op_buffer_manager import NoOpBufferManagerWorker


class AbstractBufferManager(ConfigUpdateListener, ABC):
    def __init__(self, config_manager):
        self.__config_manager: ConfigManager = config_manager
        self.__dummy_worker: BufferManagerWorker = NoOpBufferManagerWorker()
        self.__worker: Optional[BufferManagerWorker] = self.__dummy_worker

    @abstractmethod
    def create_worker(self, agent_config: AgentConfig) -> BufferManagerWorker:
        pass

    @property
    def worker(self):
        return self.__worker

    def on_successful_config_update(self, agent_config: AgentConfig):
        if self.is_refresh_needed(self.__worker.get_operating_config(), agent_config):
            old_worker = self.__worker
            self.__worker = self.create_worker(agent_config)
            old_worker.shutdown()

    def on_erroneous_config_update(self):
        old_worker = self.__worker
        self.__worker = self.__dummy_worker
        old_worker.shutdown()

    def init(self):
        self.__config_manager.subscribe_to_updates(self)

    def shutdown(self):
        if self.__worker:
            self.__worker.shutdown()
            self.__worker = None
        return True

    @staticmethod
    def is_refresh_needed(older_config: AgentConfig, new_config: AgentConfig) -> bool:
        if new_config.timestamp.timestamp() == 0 and older_config.timestamp.timestamp() == 0:
            return False
        return new_config.timestamp > older_config.timestamp



from typing import List

from apipulse_python.sdk_logger import logger


class SdkOptions:
    def __init__(
        self,
        url: str,
        application_name: str,
        auth_key: str,
        environment: str,
        logging_enabled: bool,
        log_level: str,
        mask_headers: List[str],
        capture: str,
        partner_id: str,
        team_name: str,
    ):
        self.url: str = url
        self.application_name: str = application_name
        self.auth_key: str = auth_key
        self.environment: str = environment
        self.logging_enabled: bool = logging_enabled
        self.log_level: str = log_level
        self.mask_headers: List[str] = mask_headers
        self.capture: str = capture
        self.partner_id = partner_id
        self.team_name = team_name

    def sanitize(self):
        self.url = self.url if not self.url else self.url.strip()
        self.application_name = self.application_name if not self.application_name else self.application_name.strip()
        self.auth_key = self.auth_key if not self.auth_key else self.auth_key.strip()
        self.environment = self.environment if not self.environment else self.environment.strip()
        self.capture = self.capture.strip() if self.capture and self.capture.strip() == "always" else None
        self.partner_id = self.partner_id if not self.partner_id else self.partner_id.strip()
        self.team_name = self.team_name if not self.team_name else self.team_name.strip()

    def validate(self):
        if not self.environment or self.environment == "":
            logger.error("Missing option: `environment` is required")
            return False
        if self.url == "":
            logger.error("Missing option: `url` is required")
            return False
        if self.application_name == "":
            logger.error("Missing option: `application_name` is required")
            return False
        if not self.partner_id:
            logger.error("Missing option: `partner_id` is required")
            return False
        if not self.team_name:
            logger.error("Missing option: `team_name` is required")
            return False
        return True

    def __str__(self):
        return (
            f"SdkOptions(url={self.url}, "
            f"application_name={self.application_name}, "
            f"auth_key={self.auth_key}, "
            f"environment={self.environment}, "
            f"logging_enabled={self.logging_enabled}, "
            f"log_level={self.log_level}, "
            f"mask_headers={self.mask_headers}, "
            f"capture={self.capture}, "
            f"partner_id={self.partner_id}, "
            f"team_name={self.team_name}),"
        )



from enum import Enum
from threading import Timer
from typing import List, Optional

from apipulse_python.sdk_logger import logger

from ..http.http_connection import ApiPulseHttpConnection
from ..model.agent_config import AgentConfig
from ..model.sdk_options import SdkOptions
from .config_manager import ConfigManager
from .config_update_listener import ConfigUpdateListener
from .config_validators import is_agent_config_valid

class SimpleConfigManager(ConfigManager):
    class ConfigOrError:
        class ErrorCode(Enum):
            TIMEOUT = 1
            PARSE_ERROR = 2
            INVALID_CONFIG = 3

        def __init__(self, agent_config=None, error_code=None):
            self.agent_config: AgentConfig = agent_config
            self.error_code: SimpleConfigManager.ConfigOrError.ErrorCode = error_code

        def __str__(self) -> str:
            return f"ConfigOrError(agent_config={self.agent_config}, " f"error_code={self.error_code})"

    def __init__(self, opts: SdkOptions, http_connection, agent_id):
        self.__app_name = opts.application_name
        self.__http_connection: ApiPulseHttpConnection = http_connection
        self.__agent_id = agent_id
        self.__config_update_listeners: List[ConfigUpdateListener] = []
        self.__execution_service: Optional[Timer] = None

    def init(self) -> bool:
        try:
            self.schedule_config_refresh(60)
            return True
        except Exception as e:
            logger.error("Error in SimpleConfigManager.init", exc_info=e)
            return False

    def subscribe_to_updates(self, config_update_listener) -> bool:
        self.__config_update_listeners.append(config_update_listener)
        return True

    def shutdown(self) -> bool:
        try:
            self.__execution_service.cancel()
            return True
        except Exception as e:
            logger.error("Error while shutting down SimpleConfigManager.__execution_service", exc_info=e)
            return False

    def schedule_config_refresh(self, time_sec: int):
        if self.__execution_service and self.__execution_service.is_alive():
            self.__execution_service.cancel()
        self.__execution_service = Timer(time_sec, self.fetch_config_notify)
        self.__execution_service.daemon = True
        self.__execution_service.start()


    def fetch_config_notify(self) -> None:
        config_or_error: SimpleConfigManager.ConfigOrError = self.fetch_config()
        if config_or_error.error_code:
            self.schedule_config_refresh(60)
            self.on_unsuccessful_config_fetch()
        else:
            if config_or_error.agent_config:
                self.schedule_config_refresh(config_or_error.agent_config.config_fetch_freq_in_sec)
                self.on_successful_config_fetch(config_or_error.agent_config)

    def fetch_config(self) -> ConfigOrError:
        try:
            agent_config = self.__http_connection.agent_config(agent_id=self.__agent_id, app_name=self.__app_name)
            if agent_config is not None:
                if is_agent_config_valid(agent_config):
                    return SimpleConfigManager.ConfigOrError(agent_config)
                else:
                    logger.error("Received invalid config: %s", str(agent_config))
                    return SimpleConfigManager.ConfigOrError(
                        error_code=SimpleConfigManager.ConfigOrError.ErrorCode.INVALID_CONFIG
                    )
            else:
                return SimpleConfigManager.ConfigOrError(
                    error_code=SimpleConfigManager.ConfigOrError.ErrorCode.INVALID_CONFIG
                )
        except Exception as e:
            logger.error("Error while parsing config", exc_info=e)
            return SimpleConfigManager.ConfigOrError(error_code=SimpleConfigManager.ConfigOrError.ErrorCode.PARSE_ERROR)

    def on_successful_config_fetch(self, agent_config: AgentConfig) -> None:
        for listener in self.__config_update_listeners:
            listener.on_successful_config_update(agent_config)

    def on_unsuccessful_config_fetch(self) -> None:
        for listener in self.__config_update_listeners:
            listener.on_erroneous_config_update()



from typing import Optional

import apipulse_python.core.model as model

from .http_request import HttpRequest
from .http_response import HttpResponse


class RequestResponseContext:
    def __init__(
        self,
        request=None,
        response=None,
        application_name=None,
        observed_api=None,
        api_config=None,
        agent_config=None,
        api_buffer_key=None,
        payload_capture_attempted=None,
        request_payload_capture_attempted=None,
        response_payload_capture_attempted=None,
        latency=None,
    ):
        self.request: Optional[HttpRequest] = request
        self.response: Optional[HttpResponse] = response
        self.application_name: Optional[str] = application_name
        self.observed_api: Optional[model.ObservedApi] = observed_api
        self.api_config: Optional[model.ApiConfig] = api_config
        self.agent_config: Optional[model.AgentConfig] = agent_config
        self.api_buffer_key: Optional[any] = api_buffer_key
        self.payload_capture_attempted: Optional[bool] = payload_capture_attempted
        self.request_payload_capture_attempted: Optional[bool] = request_payload_capture_attempted
        self.response_payload_capture_attempted: Optional[bool] = response_payload_capture_attempted
        self.latency: Optional[int] = latency

    def __str__(self):
        return (
            f"RequestResponseContext(request={self.request}, response={self.response}, "
            f"application_name={self.application_name}, observed_api={self.observed_api}, "
            f"api_config={self.api_config}, agent_config={self.agent_config}, "
            f"api_buffer_key={self.api_buffer_key}, "
            f"payload_capture_attempted={self.payload_capture_attempted}, "
            f"request_payload_capture_attempted={self.request_payload_capture_attempted}, "
            f"response_payload_capture_attempted={self.response_payload_capture_attempted}, "
            f"latency={self.latency})"
        )

    def __eq__(self, other):
        if not isinstance(other, RequestResponseContext):
            return False
        return (
            self.request == other.request
            and self.response == other.response
            and self.application_name == other.application_name
            and self.observed_api == other.observed_api
            and self.api_config == other.api_config
            and self.agent_config == other.agent_config
            and self.api_buffer_key == other.api_buffer_key
            and self.payload_capture_attempted == other.payload_capture_attempted
            and self.request_payload_capture_attempted == other.request_payload_capture_attempted
            and self.response_payload_capture_attempted == other.response_payload_capture_attempted
            and self.latency == other.latency
        )



from typing import List


def get_path_segments(uri: str) -> List[str]:
    if not uri:
        return []
    return [segment for segment in uri.split("/") if len(segment) > 0]


def are_path_segments_matching(path_variable_a: str, path_variable_b: str) -> bool:
    if path_variable_a is None and path_variable_b is None:
        return True
    if path_variable_a and path_variable_b:
        if path_variable_a == path_variable_b:
            return True
        is_path_variable_a_template = is_path_segment_template(path_variable_a)
        is_path_variable_b_template = is_path_segment_template(path_variable_b)
        return is_path_variable_a_template or is_path_variable_b_template
    return False


def is_path_segment_template(path_segment: str) -> bool:
    if path_segment is None:
        return False
    path_segment = path_segment.strip()
    return path_segment.startswith("{") and path_segment.endswith("}")



from typing import List

from apipulse_python.core.util.uri_utils import (
    are_path_segments_matching,
    get_path_segments,
    is_path_segment_template,
)


class URI:
    def __init__(self, data: dict):
        self.uri_path: str = data["uriPath"]
        self.has_path_variable: bool = data["hasPathVariable"]

    @staticmethod
    def get_non_templated_uri(uri_path: str):
        return URI({"uriPath": uri_path, "hasPathVariable": False})

    @staticmethod
    def get_uri(uri_path: str):
        path_segments = get_path_segments(uri_path)
        is_templated_uri = False
        for path_segment in path_segments:
            if is_path_segment_template(path_segment):
                is_templated_uri = True
                break
        return URI({"uriPath": uri_path, "hasPathVariable": is_templated_uri})

    def __str__(self):
        return f"URI(uri_path='{self.uri_path}', has_path_variable={self.has_path_variable})"
    

    def __hash__(self):
        return hash((self.uri_path, self.has_path_variable))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, URI) or other is None:
            return False
        if not self.has_path_variable and not other.has_path_variable and self.uri_path == other.uri_path:
            return True
        path_segments = get_path_segments(self.uri_path)
        other_path_segments = get_path_segments(other.uri_path)
        if len(path_segments) != len(other_path_segments):
            return False
        for idx in range(len(path_segments)):
            path_segment = path_segments[idx]
            other_path_segment = other_path_segments[idx]
            if not are_path_segments_matching(path_segment, other_path_segment):
                return False
        return True

    def get_size(self) -> int:
        return len(get_path_segments(self.uri_path))

    def get_path_segments(self) -> List[str]:
        return get_path_segments(self.uri_path)



from threading import Semaphore

from apipulse_python.sdk_logger import logger

from ..http.http_connection import ApiPulseHttpConnection
from ..model import AgentConfig
from .abstract_buffer_manager import AbstractBufferManager
from .api_buffer_key import ApiBufferKey
from .buffer_manager_worker import BufferManagerWorker
from .simple_buffer import SimpleBuffer


class DiscoveredApiBufferManager(AbstractBufferManager):
    def __init__(self, config_manager, apipulse_http_connection):
        super().__init__(config_manager)
        self.__apipulse_http_connection = apipulse_http_connection

    def create_worker(self, agent_config: AgentConfig) -> BufferManagerWorker:
        return DiscoveredApiBufferManager.DiscoveredApiBufferManagerWorker(
            agent_config, self.__apipulse_http_connection
        )

    class DiscoveredApiBufferManagerWorker(BufferManagerWorker):
        def __init__(self, agent_config: AgentConfig, http_connection):
            super().__init__(agent_config)
            self.__semaphore = Semaphore(agent_config.discovery_buffer_size)
            self.__http_connection: ApiPulseHttpConnection = http_connection

        def init(self) -> bool:
            return True

        def offer(self, key: ApiBufferKey, sample) -> bool:
            if not self._buffer_map.get(key):
                self._buffer_map[key] = SimpleBuffer(super().get_operating_config().discovery_buffer_size_per_api)
            buffer = self._buffer_map.get(key)
            if buffer:
                return buffer.offer(sample)
            else:
                pass
            return False

        def can_offer(self, key: ApiBufferKey) -> bool:
            if self.__semaphore.acquire(blocking=False):
                can_offer = False
                if super().get_operating_config().capture_api_sample:
                    buffer = self._buffer_map.get(key)
                    if buffer:
                        can_offer = buffer.can_offer()
                    else:
                        can_offer = len(self._buffer_map) < super().get_operating_config().discovery_buffer_size
                self.__semaphore.release()
                return can_offer
            return False

        def sync_for_key(self, key: ApiBufferKey):
            try:
                buffer = self._buffer_map.get(key)
                if not buffer:
                    return
                iterations = buffer.size()
                if iterations == 0:
                    self._buffer_map.pop(key)
                    return
                contents = []
                while iterations > 0:
                    iterations -= 1
                    sample = buffer.poll()
                    if not sample:
                        self._buffer_map.pop(key)
                        break
                    contents.append(sample)
                if len(contents) == 0:
                    return
                self.__http_connection.send_samples(contents)
            except Exception as e:
                logger.error("Error inside syncForKey for key " + str(key.uri), exc_info=e)




HEADERS = None

def get_headers():
    global HEADERS
    if HEADERS:
        return HEADERS
    else:
        build_headers()
        return HEADERS


def build_headers():
    from ..apipulse_sdk import options as OPTIONS
    from apipulse_python.sdk_logger import logger
    from apipulse_python.core.model.sdk_options import SdkOptions
    options: SdkOptions = OPTIONS
    global HEADERS
    HEADERS = {
        'X-API-KEY': options.auth_key,
        'X-PARTNER-ID': options.partner_id,
        'X-ENV-NAME': options.environment,
        'X-TEAM-NAME': options.team_name,
        'X-SERVICE-NAME': options.application_name,
        'Content-Type': 'application/json'
    }



from pydantic import BaseModel
from typing import Dict, List

class EndpointInfo(BaseModel):
    pattern: str
    method: str

    # Pydantic's BaseModel includes a __str__ method that uses __repr__
    # and a method .dict() which is equivalent to your to_dict()

class ServiceApiDetails(BaseModel):
    controllerVsApis: Dict[str, List[EndpointInfo]]

    # Pydantic will automatically validate that controllerVsApis is a dict
    # with string keys and lists of EndpointInfo objects as values.
    # The .dict() method and __str__ method are also provided by BaseModel.


class ApiOwner(BaseModel):
    env: str
    team: str
    serviceName: str
    
class ApiSampleWrapper(BaseModel):
    apiOwner: ApiOwner
    apiSamples: List # it is better to mention the type of object in list
    
    


from ..constant import HttpMethod
from ..model import URI, ApiConfig, ObservedApi


class ApiBufferKey:
    def __init__(self, uri, method):
        self.uri: URI = uri
        self.method: HttpMethod = method

    @staticmethod
    def get_api_buffer_key_from_observed_api(observed_api: ObservedApi):
        return ApiBufferKey(observed_api.uri, observed_api.method)

    @staticmethod
    def get_api_buffer_key_from_api_config(api_config: ApiConfig):
        return ApiBufferKey(api_config.uri, api_config.method)

    def __hash__(self):
        return hash((self.uri, self.method))

    def __eq__(self, other):
        if not isinstance(other, ApiBufferKey):
            return False
        return self.uri == other.uri and self.method == other.method

    def __str__(self):
        return f"ApiBufferKey({self.uri}, {self.method})"



from typing import Optional

from apipulse_python.core.api_processor import ApiProcessor
from apipulse_python.core.buffer.api_buffer_key import ApiBufferKey
from apipulse_python.core.config.config_manager import ConfigManager
from apipulse_python.core.config.config_update_listener import ConfigUpdateListener
from apipulse_python.core.http.context import RequestResponseContext
from apipulse_python.core.model import AgentConfig
from apipulse_python.core.util.filter_utils import (
    get_api_config,
    get_observed_api_from_request,
    is_blacklisted_api,
)
from apipulse_python.sdk_logger import logger
import time

class ApipulseLoopFilter(ConfigUpdateListener):
    def __init__(self, config_manager, api_processor, application_name):
        self.__agent_config: Optional[AgentConfig] = None
        self.__configManager: ConfigManager = config_manager
        self.__api_processor: ApiProcessor = api_processor
        self.__application_name: str = application_name

    def init(self) -> bool:
        return self.__configManager.subscribe_to_updates(self)

    def on_successful_config_update(self, agent_config: AgentConfig):
        self.__agent_config = agent_config
        logger.debug(f"Config fetch success: {self.__agent_config}")

    def on_erroneous_config_update(self):
        self.__agent_config = AgentConfig.get_no_op_agent_config()
        logger.debug(f"Config fetch failed: {self.__agent_config}")

    def process(self, ctx: RequestResponseContext, next_fn, *args, **kwargs):
        start_time = time.time()
        framework_response, http_response = next_fn(*args, **kwargs)
        end_time = time.time()
        try:
            agent_config_local = self.__agent_config

            if not agent_config_local:
                return framework_response
            if agent_config_local.capture_api_sample is False:
                return framework_response
            if not ctx.request:
                return framework_response

            observed_api = get_observed_api_from_request(ctx.request)

            if is_blacklisted_api(observed_api, agent_config_local):
                return framework_response

            ctx.observed_api = observed_api
            ctx.application_name = self.__application_name
            ctx.agent_config = agent_config_local

            api_config = get_api_config(observed_api, agent_config_local)
            ctx.latency = int((end_time - start_time) * 1000)
            if api_config:
                logger.debug("Found sample for existing API")
                ctx.api_config = api_config
                ctx.api_buffer_key = ApiBufferKey.get_api_buffer_key_from_api_config(ctx.api_config)
                self.__api_processor.process_registered_api(ctx, http_response)
            else:
                logger.debug("Found sample for new API")
                ctx.api_buffer_key = ApiBufferKey.get_api_buffer_key_from_observed_api(ctx.observed_api)
                self.__api_processor.process_discovered_api(ctx, http_response)
        except Exception as e:
            logger.error("Error in Apipulse.process", exc_info=e)
        return framework_response



from threading import Semaphore

from apipulse_python.sdk_logger import logger

from ..http.http_connection import ApiPulseHttpConnection
from ..model import AgentConfig
from .abstract_buffer_manager import AbstractBufferManager
from .api_buffer_key import ApiBufferKey
from .buffer_manager_worker import BufferManagerWorker
from .simple_buffer import SimpleBuffer


class RegisteredApiBufferManager(AbstractBufferManager):
    def __init__(self, config_manager, apipulse_http_connection):
        super().__init__(config_manager)
        self.__apipulse_http_connection = apipulse_http_connection

    def create_worker(self, agent_config: AgentConfig) -> BufferManagerWorker:
        return RegisteredApiBufferManager.RegisteredApiBufferManagerWorker(
            agent_config, self.__apipulse_http_connection
        )

    class RegisteredApiBufferManagerWorker(BufferManagerWorker):
        def __init__(self, agent_config: AgentConfig, http_connection):
            super().__init__(agent_config)
            self.__semaphore = Semaphore(self.get_registered_api_count_to_capture())
            self.__http_connection: ApiPulseHttpConnection = http_connection

        def get_registered_api_count_to_capture(self):
            agent_config = super().get_operating_config()
            if not agent_config.registered_api_configs or len(agent_config.registered_api_configs) == 0:
                return 0

            total_apis = 0
            for api_config in agent_config.registered_api_configs:
                if api_config.buffer_size:
                    total_apis += api_config.buffer_size

            return total_apis

        def get_registered_api_buffer_size(self, api_buffer_key: ApiBufferKey):
            agent_config = super().get_operating_config()
            if not agent_config.registered_api_configs or len(agent_config.registered_api_configs) == 0:
                return 0

            for api_config in agent_config.registered_api_configs:
                if api_config.method == api_buffer_key.method and api_config.uri == api_buffer_key.uri:
                    return api_config.buffer_size

            return 0

        def init(self) -> bool:
            return True

        def offer(self, key: ApiBufferKey, sample) -> bool:
            if not self._buffer_map.get(key):
                self._buffer_map[key] = SimpleBuffer(self.get_registered_api_buffer_size(key))
            buffer = self._buffer_map.get(key)
            if buffer:
                return buffer.offer(sample)
            else:
                pass

            return False

        def can_offer(self, key: ApiBufferKey) -> bool:
            buffer_size = self.get_registered_api_buffer_size(key)
            if buffer_size == 0:
                return False

            buffer = self._buffer_map.get(key)

            if self.__semaphore.acquire(blocking=False):
                if buffer:
                    can_offer = True
                else:
                    can_offer = buffer.can_offer()
                self.__semaphore.release()
                return can_offer
            return False

        def sync_for_key(self, key: ApiBufferKey):
            try:
                buffer = self._buffer_map.get(key)
                if not buffer:
                    logger.error("Buffer is null for " + str(key.uri))
                    return
                iterations = buffer.size()
                if iterations == 0:
                    self._buffer_map.pop(key)
                    return
                while iterations > 0:
                    iterations -= 1
                    contents = []
                    sample = buffer.poll()
                    if not sample:
                        self._buffer_map.pop(key)
                        break
                    contents.append(sample)
                    self.__http_connection.send_samples(contents)
            except Exception as e:
                logger.error("Error inside syncForKey for key " + str(key.uri), exc_info=e)



import random
from abc import ABC, abstractmethod
from typing import Dict, Optional

from apipulse_python.sdk_logger import logger

from ..model import AgentConfig
from ..scheduled_timer import ScheduledTimer
from .api_buffer_key import ApiBufferKey
from .buffer import Buffer


class BufferManagerWorker(ABC):
    def __init__(self, agent_config):
        self._buffer_map: Dict[ApiBufferKey, Buffer] = {}
        self.__agent_config: AgentConfig = agent_config
        self.__buffer_sync_timer: Optional[ScheduledTimer] = ScheduledTimer(
            self.__agent_config.buffer_sync_freq_in_sec, self.sync_for_keys
        )
        self.__buffer_sync_timer.setName("sl-buffer-" + str(random.randrange(0, 1000)))
        self.__buffer_sync_timer.daemon = True
        self.__buffer_sync_timer.start()

    @abstractmethod
    def init(self) -> bool:
        pass

    @abstractmethod
    def offer(self, key: ApiBufferKey, sample) -> bool:
        pass

    @abstractmethod
    def can_offer(self, key: ApiBufferKey) -> bool:
        pass

    @abstractmethod
    def sync_for_key(self, key: ApiBufferKey):
        pass

    def get_operating_config(self):
        return self.__agent_config

    def shutdown(self):
        try:
            self.__buffer_sync_timer.cancel()
            self.clean_up_buffer_map()
        except Exception as e:
            logger.error("Error while shutting down BufferManagerWorker.__buffer_sync_timer", exc_info=e)

    def clean_up_buffer_map(self):
        self.sync_for_keys()
        for key in list(self._buffer_map.keys()):
            buffer = self._buffer_map.get(key)
            if buffer:
                buffer.clear()
        self._buffer_map.clear()

    def sync_for_keys(self):
        keys = list(self._buffer_map.keys())
        if keys and len(keys) == 0:
            return
        for key in keys:
            self.sync_for_key(key)



from ..constant import HttpMethod
from .uri import URI


class ApiConfig:
    def __init__(self, data: dict):
        self.uri: URI = URI(data["uri"])
        self.method: HttpMethod = data["method"]
        self.buffer_size: int = data["bufferSize"]
        self.capture_sample_request: bool = data["captureSampleRequest"]
        self.capture_sample_response: bool = data["captureSampleResponse"]

    def __str__(self):
        return (
            f"ApiConfig(uri={str(self.uri)}, method={self.method}, buffer_size={self.buffer_size}, "
            f"capture_sample_request={self.capture_sample_request}, "
            f"capture_sample_response={self.capture_sample_response})"
        )



from typing import Dict, List


class HttpRequest:
    def __init__(self, raw_uri, hostname, scheme, port, method, headers, params, body_raw=None, body=None):
        self.raw_uri = raw_uri
        self.hostname = hostname
        self.scheme: str = scheme
        self.port = port
        self.method: str = method
        self.headers: Dict[str, str] = headers
        self.params: Dict[str, List[str]] = params
        self.body_raw: any = body_raw
        self.body: str = body

    def __str__(self):
        return (
            f"HttpRequest(raw_uri={self.raw_uri}, hostname={self.hostname}, "
            f"scheme={self.scheme}, method={self.method}, headers={self.headers}, "
            f"params={self.params}, body_raw={self.body_raw} body={self.body})"
        )

    def __eq__(self, other):
        if not isinstance(other, HttpRequest):
            return False
        return (
            self.raw_uri == other.raw_uri
            and self.hostname == other.hostname
            and self.scheme == other.scheme
            and self.method == other.method
            and self.headers == other.headers
            and self.params == other.params
            and self.body_raw == other.body_raw
            and self.body == other.body
        )



import random
from typing import Optional

from apipulse_python.core.filter import AlwaysCaptureApipulseFilter, ApipulseLoopFilter
from apipulse_python.sdk_logger import logger, set_logging_config

from .always_capture_sync import ApiPulseDataSyncService
from .api_processor import ApiProcessor
from .buffer.discovered_buffer_manager import DiscoveredApiBufferManager
from .buffer.registered_buffer_manager import RegisteredApiBufferManager
from .config.simple_config_manager import SimpleConfigManager
from .http.http_connection import ApiPulseHttpConnection
from .model.sdk_options import SdkOptions

class ApiPulseAutoConfiguration:
    def __init__(self, opts):
        self.__opts: SdkOptions = opts
        self.__filter: Optional[ApipulseLoopFilter] = None

    def init(self):
        agent_id = str(random.randrange(100000))
        set_logging_config(self.__opts.logging_enabled, self.__opts.log_level)

        http_connection = ApiPulseHttpConnection(
            ct_url=self.__opts.url,
            auth_key=self.__opts.auth_key,
            environment=self.__opts.environment,
            capture=self.__opts.capture,
        )

        if self.__opts.capture == "always":
            data_sync_service = ApiPulseDataSyncService(apipulse_http_connection=http_connection)
            data_sync_service.init()

            api_processor = ApiProcessor(
                mask_headers=self.__opts.mask_headers,
                discovered_api_buffer_manager=None,
                registered_api_buffer_manager=None,
                data_sync_service=data_sync_service,
            )

            self.__filter = AlwaysCaptureApipulseFilter(
                api_processor=api_processor, application_name=self.__opts.application_name
            )

            logger.debug('ApipulseAutoConfiguration.__opts.capture == "always"')

            return

        config_manager = SimpleConfigManager(opts=self.__opts, agent_id=agent_id, http_connection=http_connection)
        config_manager.init()

        discovered_buffer_manager = DiscoveredApiBufferManager(
            config_manager=config_manager, apipulse_http_connection=http_connection
        )
        logger.debug("discovered_buffer_manager.init")
        discovered_buffer_manager.init()

        registered_buffer_manager = RegisteredApiBufferManager(
            config_manager=config_manager, apipulse_http_connection=http_connection
        )
        logger.debug("registered_buffer_manager.init")
        registered_buffer_manager.init()

        api_processor = ApiProcessor(
            mask_headers=self.__opts.mask_headers,
            discovered_api_buffer_manager=discovered_buffer_manager,
            registered_api_buffer_manager=registered_buffer_manager,
        )

        self.__filter = ApipulseLoopFilter(
            config_manager=config_manager, api_processor=api_processor, application_name=self.__opts.application_name
        )

        logger.debug("ApipulseAutoConfiguration.__filter.init")
        self.__filter.init()

    @property
    def filter(self) -> Optional[ApipulseLoopFilter]:
        return self.__filter



from setuptools import setup, find_packages

setup(
    name='apipulse-sdk',
    version='0.1.1',
    packages=find_packages(),
    # include_package_data=True,
    description='Api automation testing sdk for django framework',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='http://github.com/yourusername/your-middleware',
    author='Vikram Panwar',
    author_email='vikram.panwar@apipulse.dev',
    license='unlicensed',
    install_requires=[
        'Django>=3.0',
    ],
    classifiers=[
        'Framework :: Django',
        'Intended Audience :: Developers',
        "License :: Other/Proprietary License",
    ],
)



from abc import ABC, abstractmethod

from ..model import ApiSample


class Buffer(ABC):
    @abstractmethod
    def offer(self, sample: ApiSample) -> bool:
        pass

    @abstractmethod
    def can_offer(self):
        pass

    @abstractmethod
    def poll(self) -> ApiSample:
        pass

    @abstractmethod
    def clear(self):
        pass

    @abstractmethod
    def size(self):
        pass



from typing import List

from ..constant import HttpMethod
from .uri import URI


class BlackListRule:
    def __init__(self, data: dict):
        self.black_list_type: str = data["blackListType"]
        self.match_values: List[str] = data["matchValues"]
        self.method: HttpMethod = data["method"]

    def is_valid(self) -> bool:
        if not self.black_list_type or not self.match_values:
            return False
        return True

    def matches_uri(self, uri: URI, method: HttpMethod) -> bool:
        if not self.is_valid():
            return False
        if not uri or not method:
            return False
        if self.method and self.method != method:
            return False
        if self.black_list_type.lower() == "endswith":
            for match_value in self.match_values:
                if uri.uri_path.lower().endswith(match_value.lower()):
                    return True
        elif self.black_list_type.lower() == "absolute":
            for match_value in self.match_values:
                if uri == URI.get_uri(match_value):
                    return True
        return False



from ..model import AgentConfig
from .api_buffer_key import ApiBufferKey
from .buffer_manager_worker import BufferManagerWorker


class NoOpBufferManagerWorker(BufferManagerWorker):
    def __init__(self):
        super().__init__(AgentConfig.get_no_op_agent_config())

    def init(self) -> bool:
        return True

    def offer(self, key: ApiBufferKey, sample) -> bool:
        pass

    def can_offer(self, key: ApiBufferKey) -> bool:
        pass

    def sync_for_key(self, key):
        return



from .context import RequestResponseContext
from .http_connection import ApiPulseHttpConnection
from .http_request import HttpRequest
from .http_response import HttpResponse

__all__ = ["HttpRequest", "HttpResponse", "RequestResponseContext", "ApiPulseHttpConnection"]



from datetime import datetime, timezone
from typing import List

from apipulse_python.core.model.api_config import ApiConfig
from apipulse_python.core.model.black_list_rule import BlackListRule


class AgentConfig:
    def __init__(
        self,
        buffer_sync_freq_in_sec: int,
        capture_api_sample: bool,
        config_fetch_freq_in_sec: int,
        registered_api_configs: List[ApiConfig],
        timestamp: str,
        discovery_buffer_size: int,
        discovery_buffer_size_per_api: int,
        black_list_rules: List[dict],
    ):
        self.buffer_sync_freq_in_sec: int = buffer_sync_freq_in_sec
        self.capture_api_sample: bool = capture_api_sample
        self.config_fetch_freq_in_sec: int = config_fetch_freq_in_sec
        self.registered_api_configs: List[ApiConfig] = list(
            map(lambda api_config: ApiConfig(api_config), registered_api_configs)
        )
        self.timestamp: datetime = datetime.fromisoformat(timestamp)
        self.discovery_buffer_size: int = discovery_buffer_size
        self.discovery_buffer_size_per_api: int = discovery_buffer_size_per_api
        self.black_list_rules: List[BlackListRule] = list(map(lambda rule: BlackListRule(rule), black_list_rules))

    def __str__(self) -> str:
        return (
            f"AgentConfig(buffer_sync_freq_in_sec={self.buffer_sync_freq_in_sec}, "
            f"capture_api_sample={self.capture_api_sample}, "
            f"config_fetch_freq_in_sec={self.config_fetch_freq_in_sec}, "
            f"registered_api_configs={str(self.registered_api_configs)}, "
            f"timestamp={self.timestamp}, "
            f"discovery_buffer_size={self.discovery_buffer_size}, "
            f"discovery_buffer_size_per_api={self.discovery_buffer_size_per_api}, "
            f"black_list_rules={str(self.black_list_rules)})"
        )

    @staticmethod
    def get_no_op_agent_config():
        epoch = datetime(1970, 1, 1, 0, 0, 0, tzinfo=timezone.utc)
        return AgentConfig(120, False, 120, [], epoch.isoformat(), 0, 0, [])



from apipulse_python.core.model.sdk_options import SdkOptions
from apipulse_python.core.apipulse_sdk import options as OPTIONS
from apipulse_python.core.util.common_utils import get_headers

BASE_URL: str = OPTIONS.url
HEADERS: dict = get_headers()
DATA_INGESTION_URL: str = f'{BASE_URL}/api/v1/mirror/data-ingestion/code-sample/'

